hitmds <- function(MD, Y, n_dim, rate, n_cycles) {
#HITMDS High-Throughput Dimensional Scaling (HiT-MDS)
#
# Y = hitmds(D, Y, n_dim, rate, n_cycles)
#
# Embed dissimilarity matrix D into n_dim -dimensional Euclidean vector space.
#
# Arguments:
# MD - source dissimilarity matrix
# Y  - initial target point configuration, leave empty {} for auto-init
# rate - try values {0.1 1 10 100} for optimum embedding, start using rate=1
# n_cycles - value between 10 and 5000, depending on number of data vectors;
#            the more data vectors, the less cycles are necessary.
#
# Return:
# Y - non-standardized embedded points (apply stdscatter for standardization)
#
# Watch value output after each iteration. The higher, the better.
# 0 means perfect mismatch of embedded data relations with D
# 1 means perfect reconstruction, i.e. most trustful embedding result.
#
# Severals runs are advisable for selecting optimum embedding results.
#
#
#
# Author:      Marc Strickert
# Institution: Leibniz-Institute of Crop Plant Research, IPK-Gatersleben
# Time-stamp:  Thu May  3 10:53:56     2007
# Lizence:     GPLv2; not for use in critical applications
#
# Special thanks to Dr. Alexander Ihlow 
# from IPK-Gatersleben for providing vectorized MATLAB code.

  if(!is.matrix(MD))
    stop('MD no matrix!')

  if(!isSymmetric(MD))
    stop('Distance matrix MD not symmetric!')

  n_data <- nrow(MD)
  n_datm <- n_data / (n_data-1)
  n_iter <- n_cycles * n_data
  n_itr1 <- n_iter 
  n_itr2 <- n_iter / 2

  if(n_dim == 1 && length(Y) > 0 && !is.matrix(Y))
    Y <- as.matrix(Y)

  if(length(Y) > 0 && (nrow(Y) != n_data || ncol(Y) != n_dim))
    stop('Y matrix has wrong size!')

  n_elem <- n_data * (n_data-1) * .5

  # random point initialization if necessary, not very good,
  # better: random projection if available
  if(length(Y) == 0)
    Y <- 2 * (matrix(runif(n_data*n_dim), n_data, n_dim) - .5)

  MT <- as.matrix(dist(Y, diag=T, upper=T))

  Y <- t(Y) # because of point accessing stragtegy

  cnt <- 1:n_data
  eps <- 1e-16
  
  mn_T <- n_datm * mean(MT)
  tmp <- MT - mn_T
  MD <- MD - n_datm * mean(MD)
  mi_T <- .5 * (sum(tmp *  MD) - sum(diag(tmp) * diag(MD)))
  mo_T <- .5 * (sum(tmp * tmp) - sum(diag(tmp) * diag(tmp)))
  mo_D <- .5 * (sum( MD *  MD) - sum(diag( MD) * diag(MD)))

  i_rep <- 0

  while(n_iter > 0) {

    i_rep <- i_rep - 1

    if(i_rep <= 0) {
      cat(paste(n_cycles,sqrt(max(mi_T*mi_T,eps)/(mo_D*mo_T)),sep=" "),sep="\n")
# plot(Y[1,],Y[2,],'p')
      n_cycles <- n_cycles - 1
      i_rep <- n_data
      shuffl <- sample(cnt)
    }

    # draw random index
    i <- shuffl[i_rep]
    pnt <- Y[,i]
    sq <- cnt[-i]


    # calc point i update strength
    tmp <- (MT[sq,i]-mn_T) * mi_T - MD[sq,i] * mo_T
    tmp <- tmp / pmax(MT[sq,i], eps) * t(pnt - Y[,sq]) 
    pnt_del <- apply(tmp, 2, sum)
    rt <- ifelse(n_iter<n_itr2, rate * (1 + 2 * (n_iter-n_itr2)/n_itr1), rate)
    cat(rt)
    pnt_del <- pnt - rt * pnt_del / pmax(abs(pnt_del),eps)

    # update incremental mean
    tmp <- sqrt(apply((pnt_del-Y[,sq])^2, 2, sum))
    ddif_ <- tmp - MT[sq, i]
    dif <- sum(ddif_)

    # update incremental variance
    d_mn <- dif / n_elem
    d_of <- -d_mn - mn_T

    d_mi <- sum(MD[sq,i] * ddif_)
    d_mo <- sum(ddif_ * (ddif_ + 2 * (MT[sq,i] + d_of)))

    mn_T <- mn_T + d_mn
    mi_T <- mi_T + d_mi
    mo_T <- mo_T + d_mo + n_elem * d_mn * d_mn

    # update target distance matrix
    MT[sq,i] <- MT[sq,i] + ddif_
    MT[i,sq] <- t(MT[sq,i])

    Y[,i] <- pnt_del

    n_iter <- n_iter - 1

  } # while
 
  # output final correlation of source data and reconstructions
  cat(paste(0,sqrt(max(mi_T*mi_T,eps) / (mo_D * mo_T)),sep=" "),sep="\n")
  
  t(Y)
} # function

stdscatter <- function(y) {
#STDSCATTER - normalize shift/scale/rotation invariant scatter plots by PCA

# move to origin
  y <- (t(y) - apply(y, 2, mean))

# get PCA projection (=rotation) using left eigenvectors of SVD
  y <- crossprod(svd(y, nv = 0)$u, y)

# flip heavier (larger) second moment to the left (in general: to minus)
# and rescale dimensions by overall maximum coordinate veriance
  t(y * sign(apply(sign(y) * y * y, 1, sum))) / max(apply(y, 1, sd))
}

# data vectors are arranged as rows
# data is taken from example table in book Bio-Statisik II, Precht&Kraft
dat <- as.matrix(read.table(dataFile))

# Euclidean distance matrix (leads to embedding results, similar to PCA)
MD <- as.matrix(dist(dat,diag=T,upper=T))
# 1-correlation (Pearson) similarity matrix
#  MD <- 1 - cor(t(dat))

d_goal <- 2   # target dimension
l_rate <- .1  # learning rate
n_cyc  <- 50  # number of cycles

d_inp  <- ncol(dat)

# initialize y by random projection
y <- dat %*% (2 * (matrix(runif(d_inp * d_goal), d_inp, d_goal) - .5))

# do the work
y <- hitmds(MD, y, d_goal, l_rate, n_cyc)

# standardize HiT-MDS output
y <- stdscatter(y)

png(filename=outputFile, width=320, height=320)
plot(y[,1],y[,2],'p',pch=15,cex=0.75,col="blue",ylim=c(-5,5),xlim=c(-5,5))
text(y[,1]+.15,y[,2]-.1,y<-as.matrix(read.table(labelsFile)),col=c('blue')[y],cex=1.25,font=2)
dev.off()
